/*
Copyright (c) 2021 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#define HELP_MESSAGE                                                                                \
  "GETMESSAGETABLE v. 1.0\n"                                                                        \
  "Dump the message table from a PE file, and writes the found messages into a text file.\n\n"      \
  "Syntax:\n"                                                                                       \
  "GETMESSAGETABLE [ PE_File_Path  Text_File_Path ]\n"                                              \
  "  PE_File_Path    name of a Portable Executable file (like *.exe or *.mui)\n"                    \
  "  Text_File_Path  name of a text file where the messages are written\n\n"                        \
  "Output format:\n"                                                                                \
  "#M 0xmmmmmmmm #L 0xllll : (m = HEX value of the message ID, l = HEX value of the language ID)\n" \
  "message text\n"                                                                                  \
  "extra line break (not belonging to the message text)\n\n"                                        \
  "A list of language IDs can be downloaded on the Microsoft page:\n"                               \
  "https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-lcid/"                           \
  "\n\n"

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#undef UNICODE
#define UNICODE
#undef _WIN32_WINNT
#define _WIN32_WINNT 0x0600
#undef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>
#include <strsafe.h>

static BOOL CALLBACK ResourceCB(HINSTANCE hModule, LPCWSTR lpszType, LPCWSTR lpszName, WORD wIDLanguage, LONG_PTR lParam)
{
  static BOOL bomIsWritten = FALSE;
  const HRSRC hRes = FindResourceExW((HMODULE)hModule, lpszType, lpszName, wIDLanguage);
  if (!hRes)
    return FALSE;

  const HGLOBAL hGlob = LoadResource(hModule, hRes);
  if (!hGlob)
    return FALSE;

  const LPBYTE rawData = LockResource(hGlob);
  if (!rawData)
    return FALSE;

  const PMESSAGE_RESOURCE_DATA pMessageResourceData = (PMESSAGE_RESOURCE_DATA)rawData;
  for (PMESSAGE_RESOURCE_BLOCK it = pMessageResourceData->Blocks, end = it + pMessageResourceData->NumberOfBlocks; it < end; ++it)
  {
    PMESSAGE_RESOURCE_ENTRY pMessageResourceEntry = (PMESSAGE_RESOURCE_ENTRY)(rawData + it->OffsetToEntries);
    if ((pMessageResourceEntry->Flags & 1) && !bomIsWritten)
    {
      bomIsWritten = TRUE;
      WriteFile((HANDLE)lParam, "\xFF\xFE", 2UL, NULL, NULL);
    }

    for (DWORD msgId = it->LowId; msgId <= it->HighId; ++msgId)
    {
      if (pMessageResourceEntry->Flags & 1) // UTF-16
      {
        wchar_t id[28] = { 0 };
        if (StringCchPrintfW(id, 28U, L"#M 0x%.8lX #L 0x%.4X :\r\n", msgId, wIDLanguage) != S_OK ||
            !WriteFile((HANDLE)lParam, id, 54UL, NULL, NULL) ||
            !WriteFile((HANDLE)lParam, pMessageResourceEntry->Text, (DWORD)wcslen((wchar_t *)pMessageResourceEntry->Text) << 1, NULL, NULL) ||
            !WriteFile((HANDLE)lParam, L"\r\n", 4UL, NULL, NULL))
          return FALSE;
      }
      else // ANSI
      {
        char id[28] = { 0 };
        if (StringCchPrintfA(id, 28U, "#M 0x%.8lX #L 0x%.4X :\r\n", msgId, wIDLanguage) != S_OK ||
            !WriteFile((HANDLE)lParam, id, 27UL, NULL, NULL) ||
            !WriteFile((HANDLE)lParam, pMessageResourceEntry->Text, (DWORD)strlen((char *)pMessageResourceEntry->Text), NULL, NULL) ||
            !WriteFile((HANDLE)lParam, "\r\n", 2UL, NULL, NULL))
          return FALSE;
      }

      pMessageResourceEntry = (PMESSAGE_RESOURCE_ENTRY)((LPBYTE)pMessageResourceEntry + pMessageResourceEntry->Length);
    }
  }

  return TRUE;
}

#if defined(__GNUC__) && !defined(__cplusplus) // add these linker options: -Wl,-e,__start -nostartfiles
#  ifdef __i386
void _start(void);
void _start(void)
#  else
void __start(void);
void __start(void)
#  endif
#else
int main(void)
#endif
{
  BOOL isGood = FALSE;
  HMODULE hMod = NULL;
  HANDLE hFileOut = INVALID_HANDLE_VALUE;
  int argc = 0;
  LPWSTR *argv = CommandLineToArgvW(GetCommandLineW(), &argc);
  if (!argv)
    exit(1);

  if (argc != 3)
  {
    fputs(HELP_MESSAGE, stderr);
    goto CleanUp;
  }

  hMod = LoadLibraryExW(argv[1], NULL, DONT_RESOLVE_DLL_REFERENCES | LOAD_LIBRARY_AS_DATAFILE);
  if (!hMod)
    goto CleanUp;

  hFileOut = CreateFileW(argv[2], GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hFileOut == INVALID_HANDLE_VALUE)
    goto CleanUp;

  isGood = EnumResourceLanguagesExW(hMod, RT_MESSAGETABLE, MAKEINTRESOURCEW(1) /* always 1 for message tables */, ResourceCB, (LONG_PTR)hFileOut, RESOURCE_ENUM_LN | RESOURCE_ENUM_MUI, (LANGID)0);

CleanUp:
  if (hMod)
    FreeLibrary(hMod);

  if (hFileOut != INVALID_HANDLE_VALUE)
    CloseHandle(hFileOut);

  if (argv)
    LocalFree(argv);

  exit(!isGood || GetLastError());
}
